from datetime import date, time

from django.db import transaction
from django.db.models.signals import m2m_changed, post_delete, post_save, pre_delete
from django.test import TransactionTestCase, override_settings

import pytest

from aleksis.apps.chronos.models import (
    Event,
    ExtraLesson,
    Lesson,
    LessonPeriod,
    LessonSubstitution,
    Room,
    Subject,
    SupervisionSubstitution,
    TimePeriod,
)
from aleksis.apps.chronos.util.change_tracker import TimetableDataChangeTracker
from aleksis.core.models import Group, Person, SchoolTerm

pytestmark = pytest.mark.django_db


@override_settings(CELERY_BROKER_URL="memory://localhost//")
class NotificationTests(TransactionTestCase):
    serialized_rollback = True

    def setUp(self):
        self.school_term = SchoolTerm.objects.create(
            date_start=date(2020, 1, 1), date_end=date(2020, 12, 31)
        )

        self.teacher_a = Person.objects.create(
            first_name="Teacher", last_name="A", short_name="A", email="test@example.org"
        )
        self.teacher_b = Person.objects.create(
            first_name="Teacher", last_name="B", short_name="B", email="test@example.org"
        )

        self.student_a = Person.objects.create(
            first_name="Student", last_name="A", email="test@example.org"
        )
        self.student_b = Person.objects.create(
            first_name="Student", last_name="B", email="test@example.org"
        )
        self.student_c = Person.objects.create(
            first_name="Student", last_name="C", email="test@example.org"
        )
        self.student_d = Person.objects.create(
            first_name="Student", last_name="D", email="test@example.org"
        )
        self.student_e = Person.objects.create(
            first_name="Student", last_name="E", email="test@example.org"
        )

        self.group_a = Group.objects.create(
            name="Class 9a", short_name="9a", school_term=self.school_term
        )
        self.group_a.owners.add(self.teacher_a)
        self.group_a.members.add(self.student_a, self.student_b, self.student_c)
        self.group_b = Group.objects.create(
            name="Class 9b", short_name="9b", school_term=self.school_term
        )
        self.group_b.owners.add(self.teacher_b)
        self.group_b.members.add(self.student_c, self.student_d, self.student_e)

        self.time_period_a = TimePeriod.objects.create(
            weekday=0, period=1, time_start=time(8, 0), time_end=time(9, 0)
        )
        self.time_period_b = TimePeriod.objects.create(
            weekday=1, period=2, time_start=time(9, 0), time_end=time(10, 0)
        )

        self.subject_a = Subject.objects.create(name="English", short_name="En")
        self.subject_b = Subject.objects.create(name="Deutsch", short_name="De")

        self.room_a = Room.objects.create(short_name="004", name="Room 0.04")
        self.room_b = Room.objects.create(short_name="005", name="Room 0.05")

        self.lesson = Lesson.objects.create(subject=self.subject_a)
        self.lesson.groups.set([self.group_a])
        self.lesson.teachers.set([self.teacher_a])

        self.period_1 = LessonPeriod.objects.create(
            period=self.time_period_a, room=self.room_a, lesson=self.lesson
        )
        self.period_2 = LessonPeriod.objects.create(
            period=self.time_period_b, room=self.room_a, lesson=self.lesson
        )

    def _parse_receivers(self, receivers):
        return [str(r[1]) for r in receivers]

    def test_signal_registration(self):
        for model in [Event, LessonSubstitution, ExtraLesson, SupervisionSubstitution]:
            assert "TimetableDataChangeTracker._handle_save" not in "".join(
                [str(r) for r in post_save._live_receivers(model)]
            )

        for model in [Event, LessonSubstitution, ExtraLesson, SupervisionSubstitution]:
            assert "TimetableDataChangeTracker._handle_delete" not in "".join(
                [str(r) for r in post_delete._live_receivers(model)]
            )

        assert "TimetableDataChangeTracker._handle_m2m_changed" not in "".join(
            [str(r) for r in m2m_changed._live_receivers(LessonSubstitution.teachers.through)]
        )
        assert "TimetableDataChangeTracker._handle_m2m_changed" not in "".join(
            [str(r) for r in m2m_changed._live_receivers(Event.teachers.through)]
        )
        assert "TimetableDataChangeTracker._handle_m2m_changed" not in "".join(
            [str(r) for r in m2m_changed._live_receivers(Event.groups.through)]
        )
        assert "TimetableDataChangeTracker._handle_m2m_changed" not in "".join(
            [str(r) for r in m2m_changed._live_receivers(ExtraLesson.teachers.through)]
        )
        assert "TimetableDataChangeTracker._handle_m2m_changed" not in "".join(
            [str(r) for r in m2m_changed._live_receivers(ExtraLesson.groups.through)]
        )
        assert "TimetableDataChangeTracker._handle_m2m_changed" not in "".join(
            [str(r) for r in m2m_changed._live_receivers(ExtraLesson.groups.through)]
        )

        with transaction.atomic():
            tracker = TimetableDataChangeTracker()

            for model in [Event, LessonSubstitution, ExtraLesson, SupervisionSubstitution]:
                assert "TimetableDataChangeTracker._handle_save" in "".join(
                    [str(r) for r in post_save._live_receivers(model)]
                )

            for model in [Event, LessonSubstitution, ExtraLesson, SupervisionSubstitution]:
                assert "TimetableDataChangeTracker._handle_delete" in "".join(
                    [str(r) for r in pre_delete._live_receivers(model)]
                )

            assert "TimetableDataChangeTracker._handle_m2m_changed" in "".join(
                [str(r) for r in m2m_changed._live_receivers(LessonSubstitution.teachers.through)]
            )
            assert "TimetableDataChangeTracker._handle_m2m_changed" in "".join(
                [str(r) for r in m2m_changed._live_receivers(Event.teachers.through)]
            )
            assert "TimetableDataChangeTracker._handle_m2m_changed" in "".join(
                [str(r) for r in m2m_changed._live_receivers(Event.groups.through)]
            )
            assert "TimetableDataChangeTracker._handle_m2m_changed" in "".join(
                [str(r) for r in m2m_changed._live_receivers(ExtraLesson.teachers.through)]
            )
            assert "TimetableDataChangeTracker._handle_m2m_changed" in "".join(
                [str(r) for r in m2m_changed._live_receivers(ExtraLesson.groups.through)]
            )

    def test_outside_transaction(self):
        with pytest.raises(RuntimeError):
            TimetableDataChangeTracker()

    def test_create_detection(self):

        with transaction.atomic():
            tracker = TimetableDataChangeTracker()

            assert not tracker.changes

            lesson_substitution = LessonSubstitution.objects.create(
                week=20, year=2020, lesson_period=self.period_1, cancelled=True
            )

            assert tracker.changes

            assert len(tracker.changes) == 1
            change = tracker.changes[tracker.get_instance_key(lesson_substitution)]
            assert change.instance == lesson_substitution
            assert change.created
            assert not change.deleted
            assert not change.changed_fields

            lesson_substitution.cancelled = False
            lesson_substitution.subject = self.subject_b
            lesson_substitution.save()

            assert len(tracker.changes) == 1
            change = tracker.changes[tracker.get_instance_key(lesson_substitution)]
            assert change.instance == lesson_substitution
            assert change.created
            assert not change.deleted
            assert change.changed_fields

    def test_change_detection(self):

        with transaction.atomic():
            lesson_substitution = LessonSubstitution.objects.create(
                week=20, year=2020, lesson_period=self.period_1, cancelled=True
            )

            tracker = TimetableDataChangeTracker()

            assert not tracker.changes

            lesson_substitution.cancelled = False
            lesson_substitution.subject = self.subject_b
            lesson_substitution.save()

            assert len(tracker.changes) == 1
            change = tracker.changes[tracker.get_instance_key(lesson_substitution)]
            assert change.instance == lesson_substitution
            assert not change.created
            assert not change.deleted
            assert set(change.changed_fields.keys()) == {"cancelled", "subject_id"}

            assert change.changed_fields["cancelled"]
            assert change.changed_fields["subject_id"] is None

            lesson_substitution.teachers.add(self.teacher_a)

            assert len(tracker.changes) == 1
            change = tracker.changes[tracker.get_instance_key(lesson_substitution)]
            assert change.instance == lesson_substitution
            assert not change.created
            assert not change.deleted
            assert set(change.changed_fields.keys()) == {"cancelled", "subject_id", "teachers"}

            assert change.changed_fields["teachers"] == []

            lesson_substitution.teachers.remove(self.teacher_a)

            assert len(tracker.changes) == 1
            change = tracker.changes[tracker.get_instance_key(lesson_substitution)]
            assert change.instance == lesson_substitution
            assert not change.created
            assert not change.deleted
            assert set(change.changed_fields.keys()) == {"cancelled", "subject_id", "teachers"}

            assert change.changed_fields["teachers"] == []

        with transaction.atomic():
            lesson_substitution.teachers.add(self.teacher_a)

            tracker = TimetableDataChangeTracker()

            lesson_substitution.teachers.remove(self.teacher_a)

            assert len(tracker.changes) == 1
            change = tracker.changes[tracker.get_instance_key(lesson_substitution)]
            assert change.instance == lesson_substitution
            assert not change.created
            assert not change.deleted
            assert set(change.changed_fields.keys()) == {"teachers"}

            assert change.changed_fields["teachers"] == [self.teacher_a]

    def test_delete_detected(self):
        lesson_substitution = LessonSubstitution.objects.create(
            week=20, year=2020, lesson_period=self.period_1, cancelled=True
        )

        with transaction.atomic():
            tracker = TimetableDataChangeTracker()

            pk = lesson_substitution.pk

            assert not tracker.changes

            lesson_substitution.delete()

            assert len(tracker.changes) == 1
            change = tracker.changes[f"lessonsubstitution_{pk}"]
            assert change.instance == lesson_substitution
            assert not change.created
            assert change.deleted
            assert not change.changed_fields



import zoneinfo
from datetime import datetime, timedelta
from typing import Union
from urllib.parse import urljoin

from django.conf import settings
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import date_format
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from aleksis.core.models import Notification, Person
from aleksis.core.util.core_helpers import get_site_preferences

from ..models import Event, ExtraLesson, LessonSubstitution, SupervisionSubstitution


def send_notifications_for_object(
    instance: Union[ExtraLesson, LessonSubstitution, Event, SupervisionSubstitution]
):
    """Send notifications for a change object."""
    recipients = []
    if isinstance(instance, LessonSubstitution):
        recipients += instance.lesson_period.lesson.teachers.all()
        recipients += instance.teachers.all()
        recipients += Person.objects.filter(
            member_of__in=instance.lesson_period.lesson.groups.all()
        )
    elif isinstance(instance, Event):
        recipients += instance.teachers.all()
        recipients += Person.objects.filter(member_of__in=instance.groups.all())
    elif isinstance(instance, ExtraLesson):
        recipients += instance.teachers.all()
        recipients += Person.objects.filter(member_of__in=instance.groups.all())
    elif isinstance(instance, SupervisionSubstitution):
        recipients.append(instance.teacher)
        recipients.append(instance.supervision.teacher)

    description = ""
    if isinstance(instance, LessonSubstitution):
        # Date, lesson, subject
        subject = instance.lesson_period.lesson.subject
        day = instance.date
        period = instance.lesson_period.period

        if instance.cancelled:
            description += (
                _(
                    "The {subject} lesson in the {period}. period on {day} has been cancelled."
                ).format(subject=subject.name, period=period.period, day=date_format(day))
                + " "
            )
        else:
            description += (
                _(
                    "The {subject} lesson in the {period}. period "
                    "on {day} has some current changes."
                ).format(subject=subject.name, period=period.period, day=date_format(day))
                + " "
            )

            if instance.teachers.all():
                description += (
                    ngettext(
                        "The teacher {old} is substituted by {new}.",
                        "The teachers {old} are substituted by {new}.",
                        instance.teachers.count(),
                    ).format(
                        old=instance.lesson_period.lesson.teacher_names,
                        new=instance.teacher_names,
                    )
                    + " "
                )

            if instance.subject:
                description += (
                    _("The subject is changed to {subject}.").format(subject=instance.subject.name)
                    + " "
                )

            if instance.room:
                description += (
                    _("The lesson is moved from {old} to {new}.").format(
                        old=instance.lesson_period.room.name,
                        new=instance.room.name,
                    )
                    + " "
                )

        if instance.comment:
            description += (
                _("There is an additional comment: {comment}.").format(comment=instance.comment)
                + " "
            )

    elif isinstance(instance, Event):
        if instance.date_start != instance.date_end:
            description += (
                _(
                    "There is an event that starts on {date_start}, {period_from}. period "
                    "and ends on {date_end}, {period_to}. period:"
                ).format(
                    date_start=date_format(instance.date_start),
                    date_end=date_format(instance.date_end),
                    period_from=instance.period_from.period,
                    period_to=instance.period_to.period,
                )
                + "\n"
            )
        else:
            description += (
                _(
                    "There is an event on {date} from the "
                    "{period_from}. period to the {period_to}. period:"
                ).format(
                    date=date_format(instance.date_start),
                    period_from=instance.period_from.period,
                    period_to=instance.period_to.period,
                )
                + "\n"
            )

        if instance.groups.all():
            description += _("Groups: {groups}").format(groups=instance.group_names) + "\n"
        if instance.teachers.all():
            description += _("Teachers: {teachers}").format(teachers=instance.teacher_names) + "\n"
        if instance.rooms.all():
            description += (
                _("Rooms: {rooms}").format(
                    rooms=", ".join([room.name for room in instance.rooms.all()])
                )
                + "\n"
            )
    elif isinstance(instance, ExtraLesson):
        description += (
            _("There is an extra lesson on {date} in the {period}. period:").format(
                date=date_format(instance.date),
                period=instance.period.period,
            )
            + "\n"
        )

        if instance.groups.all():
            description += _("Groups: {groups}").format(groups=instance.group_names) + "\n"
        if instance.room:
            description += _("Subject: {subject}").format(subject=instance.subject.name) + "\n"
        if instance.teachers.all():
            description += _("Teachers: {teachers}").format(teachers=instance.teacher_names) + "\n"
        if instance.room:
            description += _("Room: {room}").format(room=instance.room.name) + "\n"
        if instance.comment:
            description += _("Comment: {comment}.").format(comment=instance.comment) + "\n"
    elif isinstance(instance, SupervisionSubstitution):
        description += _(
            "The supervision of {old} on {date} between the {period_from}. period "
            "and the {period_to}. period in the area {area} is substituted by {new}."
        ).format(
            old=instance.supervision.teacher.full_name,
            date=date_format(instance.date),
            period_from=instance.supervision.break_item.after_period_number,
            period_to=instance.supervision.break_item.before_period_number,
            area=instance.supervision.area.name,
            new=instance.teacher.full_name,
        )

    day = instance.date if hasattr(instance, "date") else instance.date_start

    url = urljoin(
        settings.BASE_URL,
        reverse(
            "my_timetable_by_date",
            args=[day.year, day.month, day.day],
        ),
    )

    dt_start, dt_end = instance.time_range
    dt_start = dt_start.replace(tzinfo=zoneinfo.ZoneInfo(settings.TIME_ZONE))
    dt_end = dt_end.replace(tzinfo=zoneinfo.ZoneInfo(settings.TIME_ZONE))

    send_time = get_site_preferences()["chronos__time_for_sending_notifications"]
    number_of_days = get_site_preferences()["chronos__days_in_advance_notifications"]

    start_range = timezone.now().replace(hour=send_time.hour, minute=send_time.minute)
    if timezone.now().time() > send_time:
        start_range = start_range - timedelta(days=1)
    end_range = start_range + timedelta(days=number_of_days)

    if dt_start < start_range and dt_end < end_range:
        # Skip this, because the change is in the past
        return

    if dt_start <= end_range and dt_end >= start_range:
        # Send immediately
        send_at = timezone.now()
    else:
        # Schedule for later
        send_at = datetime.combine(
            dt_start.date() - timedelta(days=number_of_days), send_time
        ).replace(tzinfo=zoneinfo.ZoneInfo(settings.TIME_ZONE))

    for recipient in recipients:
        if recipient.preferences["chronos__send_notifications"]:
            n = Notification(
                recipient=recipient,
                sender=_("Timetable"),
                title=_("There are current changes to your timetable."),
                description=description,
                link=url,
                send_at=send_at,
            )
            n.save()



from typing import Sequence

from django.db.models import Count, Q
from django.forms import RadioSelect
from django.utils.translation import gettext as _

from django_filters import BooleanFilter, FilterSet, ModelMultipleChoiceFilter
from django_select2.forms import ModelSelect2MultipleWidget
from material import Layout, Row

from aleksis.core.models import Group, Person, SchoolTerm

from .models import Break, Room, Subject, SupervisionArea, TimePeriod


class MultipleModelMultipleChoiceFilter(ModelMultipleChoiceFilter):
    """Filter for filtering multiple fields with one input.

    >>> multiple_filter = MultipleModelMultipleChoiceFilter(["room", "substitution_room"])
    """

    def filter(self, qs, value):  # noqa
        if not value:
            return qs

        if self.is_noop(qs, value):
            return qs

        q = Q()
        for v in set(value):
            if v == self.null_value:
                v = None
            for field in self.lookup_fields:
                q = q | Q(**{field: v})

            qs = self.get_method(qs)(q)

        return qs.distinct() if self.distinct else qs

    def __init__(self, lookup_fields: Sequence[str], *args, **kwargs):
        self.lookup_fields = lookup_fields
        super().__init__(self, *args, **kwargs)


class LessonPeriodFilter(FilterSet):
    period = ModelMultipleChoiceFilter(queryset=TimePeriod.objects.all())
    lesson__groups = ModelMultipleChoiceFilter(
        queryset=Group.objects.all(),
        widget=ModelSelect2MultipleWidget(
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            search_fields=[
                "name__icontains",
                "short_name__icontains",
            ],
        ),
    )
    room = MultipleModelMultipleChoiceFilter(
        ["room", "current_substitution__room"],
        queryset=Room.objects.all(),
        label=_("Room"),
        widget=ModelSelect2MultipleWidget(
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            search_fields=[
                "name__icontains",
                "short_name__icontains",
            ],
        ),
    )
    lesson__teachers = MultipleModelMultipleChoiceFilter(
        ["lesson__teachers", "current_substitution__teachers"],
        queryset=Person.objects.annotate(
            lessons_count=Count(
                "lessons_as_teacher",
                filter=Q(lessons_as_teacher__validity__school_term=SchoolTerm.current)
                if SchoolTerm.current
                else Q(),
            )
        )
        .filter(lessons_count__gt=0)
        .order_by("short_name", "last_name"),
        label=_("Teachers"),
        widget=ModelSelect2MultipleWidget(
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            search_fields=[
                "first_name__icontains",
                "last_name__icontains",
                "short_name__icontains",
            ],
        ),
    )
    lesson__subject = MultipleModelMultipleChoiceFilter(
        ["lesson__subject", "current_substitution__subject"],
        queryset=Subject.objects.all(),
        label=_("Subject"),
        widget=ModelSelect2MultipleWidget(
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            search_fields=[
                "name__icontains",
                "short_name__icontains",
            ],
        ),
    )
    substituted = BooleanFilter(
        field_name="current_substitution",
        label=_("Substitution status"),
        lookup_expr="isnull",
        exclude=True,
        widget=RadioSelect(
            choices=[
                ("", _("All lessons")),
                (True, _("Substituted")),
                (False, _("Not substituted")),
            ]
        ),
    )

    def __init__(self, *args, **kwargs):
        weekday = kwargs.pop("weekday")
        super().__init__(*args, **kwargs)
        self.filters["period"].queryset = TimePeriod.objects.filter(weekday=weekday)
        self.form.layout = Layout(
            Row("period", "lesson__groups", "room"),
            Row("lesson__teachers", "lesson__subject", "substituted"),
        )


class SupervisionFilter(FilterSet):
    break_item = ModelMultipleChoiceFilter(queryset=Break.objects.all())
    area = ModelMultipleChoiceFilter(queryset=SupervisionArea.objects.all())
    teacher = MultipleModelMultipleChoiceFilter(
        ["teacher", "current_substitution__teacher"],
        queryset=Person.objects.annotate(
            lessons_count=Count(
                "lessons_as_teacher",
                filter=Q(lessons_as_teacher__validity__school_term=SchoolTerm.current)
                if SchoolTerm.current
                else Q(),
            )
        )
        .filter(lessons_count__gt=0)
        .order_by("short_name", "last_name"),
        label=_("Teacher"),
        widget=ModelSelect2MultipleWidget(
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            search_fields=[
                "first_name__icontains",
                "last_name__icontains",
                "short_name__icontains",
            ],
        ),
    )
    substituted = BooleanFilter(
        field_name="current_substitution",
        label=_("Substitution status"),
        lookup_expr="isnull",
        exclude=True,
        widget=RadioSelect(
            choices=[
                ("", _("All supervisions")),
                (True, _("Substituted")),
                (False, _("Not substituted")),
            ]
        ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filters["break_item"].queryset = Break.objects.filter(supervisions__in=self.queryset)
        self.form.layout = Layout(
            Row("break_item", "area"),
            Row("teacher", "substituted"),
        )



# Generated by Django 3.2.4 on 2021-08-05 14:20

import aleksis.core.managers
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('resint', '0006_livedocument'),
        ('reversion', '0001_squashed_0004_auto_20160611_1202'),
        ('chronos', '0008_unique_constraints'),
    ]

    operations = [
        migrations.CreateModel(
            name='AutomaticPlan',
            fields=[
                ('livedocument_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='resint.livedocument')),
                ('number_of_days', models.PositiveIntegerField(default=1, validators=[django.core.validators.MinValueValidator(1)], verbose_name='Number of days shown in the plan')),
                ('show_header_box', models.BooleanField(default=True, help_text='The header box shows affected teachers/groups.', verbose_name='Show header box')),
                ('last_substitutions_revision', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='reversion.revision', verbose_name='Revision which triggered the last update')),
            ],
            options={
                'verbose_name': 'Automatic plan',
                'verbose_name_plural': 'Automatic plans',
            },
            bases=('resint.livedocument',),
            managers=[
                ('objects', aleksis.core.managers.PolymorphicCurrentSiteManager()),
            ],
        ),
    ]



from datetime import datetime
from typing import Optional

from django.apps import apps
from django.db.models import FilteredRelation, Q
from django.http import HttpRequest, HttpResponse, HttpResponseNotFound
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import ugettext as _
from django.views.decorators.cache import never_cache

import reversion
from django_tables2 import RequestConfig
from rules.contrib.views import permission_required

from aleksis.core.models import Announcement
from aleksis.core.util import messages
from aleksis.core.util.core_helpers import has_person
from aleksis.core.util.pdf import render_pdf

from .filters import LessonPeriodFilter, SupervisionFilter
from .forms import LessonSubstitutionForm, SupervisionSubstitutionForm
from .managers import TimetableType
from .models import Holiday, LessonPeriod, Supervision, TimePeriod
from .tables import LessonsTable, SupervisionsTable
from .util.build import build_timetable, build_weekdays
from .util.change_tracker import TimetableDataChangeTracker
from .util.chronos_helpers import (
    get_classes,
    get_el_by_pk,
    get_rooms,
    get_substitution_by_id,
    get_substitutions_context_data,
    get_supervision_substitution_by_id,
    get_teachers,
)
from .util.date import CalendarWeek, get_weeks_for_year, week_weekday_to_date
from .util.js import date_unix


@permission_required("chronos.view_timetable_overview_rule")
def all_timetables(request: HttpRequest) -> HttpResponse:
    """View all timetables for persons, groups and rooms."""
    context = {}

    user = request.user
    teachers, classes, rooms = get_teachers(user), get_classes(user), get_rooms(user)

    context["teachers"] = teachers
    context["classes"] = classes
    context["rooms"] = rooms

    return render(request, "chronos/all.html", context)


@permission_required("chronos.view_my_timetable_rule")
def my_timetable(
    request: HttpRequest,
    year: Optional[int] = None,
    month: Optional[int] = None,
    day: Optional[int] = None,
) -> HttpResponse:
    """View personal timetable on a specified date."""
    context = {}

    if day:
        wanted_day = timezone.datetime(year=year, month=month, day=day).date()
        wanted_day = TimePeriod.get_next_relevant_day(wanted_day)
    else:
        wanted_day = TimePeriod.get_next_relevant_day(timezone.now().date(), datetime.now().time())

    wanted_week = CalendarWeek.from_date(wanted_day)

    if has_person(request.user):
        person = request.user.person
        type_ = person.timetable_type

        # Build timetable
        timetable = build_timetable("person", person, wanted_day)
        week_timetable = build_timetable("person", person, wanted_week)

        if type_ is None:
            # If no student or teacher, redirect to all timetables
            return redirect("all_timetables")

        super_el = person.timetable_object

        context["timetable"] = timetable
        context["week_timetable"] = week_timetable
        context["holiday"] = Holiday.on_day(wanted_day)
        context["super"] = {"type": type_, "el": super_el}
        context["type"] = type_
        context["day"] = wanted_day
        context["today"] = timezone.now().date()
        context["week"] = wanted_week
        context["periods"] = TimePeriod.get_times_dict()
        context["smart"] = True
        context["announcements"] = (
            Announcement.for_timetables().on_date(wanted_day).for_person(person)
        )
        context["week_announcements"] = (
            Announcement.for_timetables()
            .within_days(wanted_week[0], wanted_week[6])
            .for_person(person)
        )
        context["weekdays"] = build_weekdays(TimePeriod.WEEKDAY_CHOICES, wanted_week)
        context["weekdays_short"] = build_weekdays(TimePeriod.WEEKDAY_CHOICES_SHORT, wanted_week)
        context["url_prev"], context["url_next"] = TimePeriod.get_prev_next_by_day(
            wanted_day, "my_timetable_by_date"
        )

        return render(request, "chronos/my_timetable.html", context)
    else:
        return redirect("all_timetables")


@permission_required("chronos.view_timetable_rule", fn=get_el_by_pk)
def timetable(
    request: HttpRequest,
    type_: str,
    pk: int,
    year: Optional[int] = None,
    week: Optional[int] = None,
    regular: Optional[str] = None,
    is_print: bool = False,
) -> HttpResponse:
    """View a selected timetable for a person, group or room."""
    context = {}

    is_smart = regular != "regular"

    if is_print:
        is_smart = False

    el = get_el_by_pk(request, type_, pk, prefetch=True)

    if type(el) == HttpResponseNotFound:
        return HttpResponseNotFound()

    type_ = TimetableType.from_string(type_)

    if year and week:
        wanted_week = CalendarWeek(year=year, week=week)
    else:
        wanted_week = TimePeriod.get_relevant_week_from_datetime()

    # Build timetable
    timetable = build_timetable(type_, el, wanted_week, with_holidays=is_smart)
    context["timetable"] = timetable

    # Add time periods
    context["periods"] = TimePeriod.get_times_dict()

    # Build lists with weekdays and corresponding dates (long and short variant)
    context["weekdays"] = build_weekdays(
        TimePeriod.WEEKDAY_CHOICES, wanted_week, with_holidays=is_smart
    )
    context["weekdays_short"] = build_weekdays(
        TimePeriod.WEEKDAY_CHOICES_SHORT, wanted_week, with_holidays=is_smart
    )

    context["weeks"] = get_weeks_for_year(year=wanted_week.year)
    context["week"] = wanted_week
    context["type"] = type_
    context["pk"] = pk
    context["el"] = el
    context["smart"] = is_smart
    context["week_select"] = {
        "year": wanted_week.year,
        "dest": reverse(
            "timetable_by_week",
            args=[type_.value, pk, wanted_week.year, wanted_week.week],
        )[::-1]
        .replace(str(wanted_week.week)[::-1], "cw"[::-1], 1)
        .replace(str(wanted_week.year)[::-1], "year"[::-1], 1)[::-1],
    }

    if is_smart:
        start = wanted_week[TimePeriod.weekday_min]
        stop = wanted_week[TimePeriod.weekday_max]
        context["announcements"] = (
            Announcement.for_timetables().relevant_for(el).within_days(start, stop)
        )

    week_prev = wanted_week - 1
    week_next = wanted_week + 1

    context["url_prev"] = reverse(
        "timetable_by_week", args=[type_.value, pk, week_prev.year, week_prev.week]
    )
    context["url_next"] = reverse(
        "timetable_by_week", args=[type_.value, pk, week_next.year, week_next.week]
    )

    if apps.is_installed("aleksis.apps.alsijil"):
        context["is_alsijil_installed"] = True

    if is_print:
        context["back_url"] = reverse(
            "timetable_by_week",
            args=[type_.value, pk, wanted_week.year, wanted_week.week],
        )
        return render_pdf(request, "chronos/timetable_print.html", context)
    else:
        return render(request, "chronos/timetable.html", context)


@permission_required("chronos.view_lessons_day_rule")
def lessons_day(
    request: HttpRequest,
    year: Optional[int] = None,
    month: Optional[int] = None,
    day: Optional[int] = None,
) -> HttpResponse:
    """View all lessons taking place on a specified day."""
    context = {}

    if day:
        wanted_day = timezone.datetime(year=year, month=month, day=day).date()
        wanted_day = TimePeriod.get_next_relevant_day(wanted_day)
    else:
        wanted_day = TimePeriod.get_next_relevant_day(timezone.now().date(), datetime.now().time())

    # Get lessons
    lesson_periods = LessonPeriod.objects.all()

    # Get filter
    lesson_periods_filter = LessonPeriodFilter(
        request.GET,
        queryset=lesson_periods.annotate(
            current_substitution=FilteredRelation(
                "substitutions",
                condition=(
                    Q(substitutions__week=wanted_day.isocalendar()[1], substitutions__year=year)
                ),
            )
        ),
        weekday=wanted_day.weekday(),
    )
    context["lesson_periods_filter"] = lesson_periods_filter

    # Build table
    lessons_table = LessonsTable(lesson_periods_filter.qs)
    RequestConfig(request).configure(lessons_table)

    context["lessons_table"] = lessons_table
    context["day"] = wanted_day
    context["lesson_periods"] = lesson_periods

    context["datepicker"] = {
        "date": date_unix(wanted_day),
        "dest": reverse("lessons_day"),
    }

    context["url_prev"], context["url_next"] = TimePeriod.get_prev_next_by_day(
        wanted_day, "lessons_day_by_date"
    )

    return render(request, "chronos/lessons_day.html", context)


@never_cache
@permission_required("chronos.edit_substitution_rule", fn=get_substitution_by_id)
def edit_substitution(request: HttpRequest, id_: int, week: int) -> HttpResponse:
    """View a form to edit a substitution lessen."""
    context = {}

    lesson_period = get_object_or_404(LessonPeriod, pk=id_)
    wanted_week = lesson_period.lesson.get_calendar_week(week)
    context["lesson_period"] = lesson_period
    day = week_weekday_to_date(wanted_week, lesson_period.period.weekday)
    context["date"] = day

    lesson_substitution = get_substitution_by_id(request, id_, week)

    if lesson_substitution:
        edit_substitution_form = LessonSubstitutionForm(
            request.POST or None, instance=lesson_substitution
        )
    else:
        edit_substitution_form = LessonSubstitutionForm(
            request.POST or None,
        )

    context["substitution"] = lesson_substitution

    if request.method == "POST":
        if edit_substitution_form.is_valid():
            with reversion.create_revision(atomic=True):
                tracker = TimetableDataChangeTracker()

                lesson_substitution = edit_substitution_form.save(commit=False)
                if not lesson_substitution.pk:
                    lesson_substitution.lesson_period = lesson_period
                    lesson_substitution.week = wanted_week.week
                    lesson_substitution.year = wanted_week.year
                lesson_substitution.save()
                edit_substitution_form.save_m2m()

                messages.success(request, _("The substitution has been saved."))

            return redirect("lessons_day_by_date", year=day.year, month=day.month, day=day.day)

    context["edit_substitution_form"] = edit_substitution_form

    return render(request, "chronos/edit_substitution.html", context)


@permission_required("chronos.delete_substitution_rule", fn=get_substitution_by_id)
def delete_substitution(request: HttpRequest, id_: int, week: int) -> HttpResponse:
    """Delete a substitution lesson.

    Redirects back to substition list on success.
    """
    lesson_period = get_object_or_404(LessonPeriod, pk=id_)
    wanted_week = lesson_period.lesson.get_calendar_week(week)

    get_substitution_by_id(request, id_, week).delete()

    messages.success(request, _("The substitution has been deleted."))

    date = wanted_week[lesson_period.period.weekday]
    return redirect("lessons_day_by_date", year=date.year, month=date.month, day=date.day)


@permission_required("chronos.view_substitutions_rule")
def substitutions(
    request: HttpRequest,
    year: Optional[int] = None,
    month: Optional[int] = None,
    day: Optional[int] = None,
    is_print: bool = False,
) -> HttpResponse:
    """View all substitutions on a specified day."""
    context = get_substitutions_context_data(request, year, month, day, is_print)
    if not is_print:
        return render(request, "chronos/substitutions.html", context)
    else:
        return render_pdf(request, "chronos/substitutions_print.html", context)


@permission_required("chronos.view_supervisions_day_rule")
def supervisions_day(
    request: HttpRequest,
    year: Optional[int] = None,
    month: Optional[int] = None,
    day: Optional[int] = None,
) -> HttpResponse:
    """View all supervisions taking place on a specified day."""
    context = {}

    if day:
        wanted_day = timezone.datetime(year=year, month=month, day=day).date()
        wanted_day = TimePeriod.get_next_relevant_day(wanted_day)
    else:
        wanted_day = TimePeriod.get_next_relevant_day(timezone.now().date(), datetime.now().time())

    # Get supervisions
    supervisions = Supervision.objects.on_day(wanted_day).filter_by_weekday(wanted_day.weekday())

    # Get filter
    supervisions_filter = SupervisionFilter(
        request.GET,
        queryset=supervisions.annotate(
            current_substitution=FilteredRelation(
                "substitutions",
                condition=(Q(substitutions__date=wanted_day)),
            )
        ),
    )
    context["supervisions_filter"] = supervisions_filter

    # Build table
    supervisions_table = SupervisionsTable(
        supervisions_filter.qs.annotate_week(week=CalendarWeek.from_date(wanted_day))
    )
    RequestConfig(request).configure(supervisions_table)

    context["supervisions_table"] = supervisions_table
    context["day"] = wanted_day
    context["supervisions"] = supervisions

    context["datepicker"] = {
        "date": date_unix(wanted_day),
        "dest": reverse("supervisions_day"),
    }

    context["url_prev"], context["url_next"] = TimePeriod.get_prev_next_by_day(
        wanted_day, "supervisions_day_by_date"
    )

    return render(request, "chronos/supervisions_day.html", context)


@never_cache
@permission_required("chronos.edit_supervision_substitution_rule")
def edit_supervision_substitution(request: HttpRequest, id_: int, week: int) -> HttpResponse:
    """View a form to edit a supervision substitution."""
    context = {}

    supervision = get_object_or_404(Supervision, pk=id_)
    wanted_week = supervision.get_calendar_week(week)
    context["week"] = week
    context["supervision"] = supervision
    date = week_weekday_to_date(wanted_week, supervision.break_item.weekday)
    context["date"] = date

    supervision_substitution = get_supervision_substitution_by_id(request, id_, date)

    if supervision_substitution:
        edit_supervision_substitution_form = SupervisionSubstitutionForm(
            request.POST or None, instance=supervision_substitution
        )
    else:
        edit_supervision_substitution_form = SupervisionSubstitutionForm(
            request.POST or None,
        )

    context["substitution"] = supervision_substitution

    if request.method == "POST":
        if edit_supervision_substitution_form.is_valid():
            with reversion.create_revision(atomic=True):
                tracker = TimetableDataChangeTracker()

                supervision_substitution = edit_supervision_substitution_form.save(commit=False)
                if not supervision_substitution.pk:
                    supervision_substitution.supervision = supervision
                    supervision_substitution.date = date
                supervision_substitution.save()
                edit_supervision_substitution_form.save_m2m()

                messages.success(request, _("The substitution has been saved."))

            return redirect(
                "supervisions_day_by_date", year=date.year, month=date.month, day=date.day
            )

    context["edit_supervision_substitution_form"] = edit_supervision_substitution_form

    return render(request, "chronos/edit_supervision_substitution.html", context)


@permission_required("chronos.delete_supervision_substitution_rule")
def delete_supervision_substitution(request: HttpRequest, id_: int, week: int) -> HttpResponse:
    """Delete a supervision substitution.

    Redirects back to supervision list on success.
    """
    supervision = get_object_or_404(Supervision, pk=id_)
    wanted_week = supervision.get_calendar_week(week)
    date = week_weekday_to_date(wanted_week, supervision.break_item.weekday)

    get_supervision_substitution_by_id(request, id_, date).delete()

    messages.success(request, _("The substitution has been deleted."))

    return redirect("supervisions_day_by_date", year=date.year, month=date.month, day=date.day)



from typing import Any, Optional, Type

from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.db.models import Model
from django.db.models.signals import m2m_changed, post_save, pre_delete
from django.dispatch import Signal, receiver

from celery import shared_task
from reversion.models import Revision


def _get_substitution_models():
    from aleksis.apps.chronos.models import (
        Event,
        ExtraLesson,
        LessonSubstitution,
        SupervisionSubstitution,
    )

    return [LessonSubstitution, Event, ExtraLesson, SupervisionSubstitution]


class TimetableChange:
    """A change to timetable models."""

    def __init__(
        self,
        instance: Model,
        changed_fields: Optional[dict[str, Any]] = None,
        created: bool = False,
        deleted: bool = False,
    ):
        self.instance = instance
        self.changed_fields = changed_fields or {}
        self.created = created
        self.deleted = deleted


class TimetableDataChangeTracker:
    """Helper class for tracking changes in timetable models by using signals."""

    @classmethod
    def get_models(cls) -> list[Type[Model]]:
        """Return all models that should be tracked."""
        from aleksis.apps.chronos.models import (
            Event,
            ExtraLesson,
            LessonSubstitution,
            SupervisionSubstitution,
        )

        return [LessonSubstitution, Event, ExtraLesson, SupervisionSubstitution]

    def __init__(self):
        self.changes = {}
        self.m2m_fields = {}

        if transaction.get_autocommit():
            raise RuntimeError("Cannot use change tracker outside of transaction")

        for model in self.get_models():
            post_save.connect(self._handle_save, sender=model, weak=False)
            pre_delete.connect(self._handle_delete, sender=model, weak=False)

            # Register signals for all relevant m2m fields
            m2m_fields = {
                getattr(model, f.name).through: f
                for f in model._meta.get_fields()
                if f.many_to_many
            }
            self.m2m_fields.update(m2m_fields)
            for through_model, field in m2m_fields.items():
                m2m_changed.connect(self._handle_m2m_changed, sender=through_model, weak=False)

        transaction.on_commit(self.close)

    def get_instance_key(self, instance: Model) -> str:
        """Get unique string key for an instance."""
        return f"{instance._meta.model_name}_{instance.id}"

    def _add_change(self, change: TimetableChange):
        """Add a change to the list of changes and update, if necessary."""
        key = self.get_instance_key(change.instance)
        if key not in self.changes or change.deleted or change.created:
            self.changes[key] = change
        else:
            self.changes[key].changed_fields.update(change.changed_fields)

    def _handle_save(self, sender: Type[Model], instance: Model, created: bool, **kwargs):
        """Handle the save signal."""
        change = TimetableChange(instance, created=created)
        if not created:
            change.changed_fields = instance.tracker.changed()
        self._add_change(change)

    def _handle_delete(self, sender: Type[Model], instance: Model, **kwargs):
        """Handle the delete signal."""
        change = TimetableChange(instance, deleted=True)
        self._add_change(change)

    def _handle_m2m_changed(
        self,
        sender: Type[Model],
        instance: Model,
        action: str,
        model: Type[Model],
        pk_set: set,
        **kwargs,
    ):
        """Handle the m2m_changed signal."""
        if action in ["pre_add", "pre_remove", "pre_clear"]:
            field_name = self.m2m_fields[sender].name
            current_value = list(getattr(instance, field_name).all())
            if self.get_instance_key(instance) in self.changes:
                change = self.changes[self.get_instance_key(instance)]
                if field_name in change.changed_fields:
                    current_value = None

            if current_value is not None:
                change = TimetableChange(instance, changed_fields={field_name: current_value})
                self._add_change(change)

    def close(self):
        """Disconnect signals and send change signal."""
        for model in self.get_models():
            post_save.disconnect(self._handle_save, sender=model)
            pre_delete.disconnect(self._handle_delete, sender=model)

        for through_model, field in self.m2m_fields.items():
            m2m_changed.disconnect(self._handle_m2m_changed, sender=through_model)

        timetable_data_changed.send(sender=self, changes=self.changes)


chronos_revision_created = Signal()
substitutions_changed = Signal()
timetable_data_changed = Signal()


@shared_task
def handle_new_revision(revision_pk: int):
    """Handle a new revision in background using Celery."""
    revision = Revision.objects.get(pk=revision_pk)
    if revision.version_set.filter(content_type__app_label="chronos").exists():
        chronos_revision_created.send(sender=revision)


@receiver(chronos_revision_created)
def handle_substitution_change(sender: Revision, **kwargs):
    """Handle the change of a substitution-like object."""
    # Filter versions by substitution-like models
    content_types = ContentType.objects.get_for_models(*_get_substitution_models()).values()
    versions = sender.version_set.filter(content_type__in=content_types)
    if versions:
        substitutions_changed.send(sender=sender, versions=versions)



from datetime import date, datetime, timedelta
from enum import Enum
from typing import Dict, Iterable, List, Optional, Union

from django.contrib.sites.managers import CurrentSiteManager as _CurrentSiteManager
from django.db import models
from django.db.models import ExpressionWrapper, F, Func, Q, QuerySet, Value
from django.db.models.fields import DateField
from django.db.models.functions import Concat

from calendarweek import CalendarWeek

from aleksis.apps.chronos.util.date import week_weekday_from_date, week_weekday_to_date
from aleksis.core.managers import DateRangeQuerySetMixin, SchoolTermRelatedQuerySet
from aleksis.core.models import Group, Person
from aleksis.core.util.core_helpers import get_site_preferences


class ValidityRangeQuerySet(QuerySet, DateRangeQuerySetMixin):
    """Custom query set for validity ranges."""


class ValidityRangeRelatedQuerySet(QuerySet):
    """Custom query set for all models related to validity ranges."""

    def within_dates(self, start: date, end: date) -> "ValidityRangeRelatedQuerySet":
        """Filter for all objects within a date range."""
        return self.filter(validity__date_start__lte=end, validity__date_end__gte=start)

    def in_week(self, wanted_week: CalendarWeek) -> "ValidityRangeRelatedQuerySet":
        """Filter for all objects within a calendar week."""
        return self.within_dates(wanted_week[0], wanted_week[6])

    def on_day(self, day: date) -> "ValidityRangeRelatedQuerySet":
        """Filter for all objects on a certain day."""
        return self.within_dates(day, day)

    def for_validity_range(self, validity_range: "ValidityRange") -> "ValidityRangeRelatedQuerySet":
        return self.filter(validity=validity_range)

    def for_current_or_all(self) -> "ValidityRangeRelatedQuerySet":
        """Get all objects related to current validity range.

        If there is no current validity range, it will return all objects.
        """
        from aleksis.apps.chronos.models import ValidityRange

        current_validity_range = ValidityRange.current
        if current_validity_range:
            return self.for_validity_range(current_validity_range)
        else:
            return self

    def for_current_or_none(self) -> Union["ValidityRangeRelatedQuerySet", None]:
        """Get all objects related to current validity range.

        If there is no current validity range, it will return `None`.
        """
        from aleksis.apps.chronos.models import ValidityRange

        current_validity_range = ValidityRange.current
        if current_validity_range:
            return self.for_validity_range(current_validity_range)
        else:
            return None


class CurrentSiteManager(_CurrentSiteManager):
    use_in_migrations = False


class TimetableType(Enum):
    """Enum for different types of timetables."""

    GROUP = "group"
    TEACHER = "teacher"
    ROOM = "room"

    @classmethod
    def from_string(cls, s: Optional[str]):
        return cls.__members__.get(s.upper())


class LessonPeriodManager(CurrentSiteManager):
    """Manager adding specific methods to lesson periods."""

    def get_queryset(self):
        """Ensure all related lesson data is loaded as well."""
        return (
            super()
            .get_queryset()
            .select_related(
                "lesson",
                "lesson__subject",
                "period",
                "room",
                "lesson__validity",
                "lesson__validity__school_term",
            )
            .prefetch_related(
                "lesson__groups",
                "lesson__groups__parent_groups",
                "lesson__teachers",
                "substitutions",
            )
        )


class LessonSubstitutionManager(CurrentSiteManager):
    """Manager adding specific methods to lesson substitutions."""

    def get_queryset(self):
        """Ensure all related lesson data is loaded as well."""
        return (
            super()
            .get_queryset()
            .select_related(
                "lesson_period",
                "lesson_period__lesson",
                "lesson_period__lesson__subject",
                "subject",
                "lesson_period__period",
                "room",
                "lesson_period__room",
            )
            .prefetch_related(
                "lesson_period__lesson__groups",
                "lesson_period__lesson__groups__parent_groups",
                "teachers",
                "lesson_period__lesson__teachers",
            )
        )


class SupervisionManager(CurrentSiteManager):
    """Manager adding specific methods to supervisions."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return (
            super()
            .get_queryset()
            .select_related(
                "teacher",
                "area",
                "break_item",
                "break_item__after_period",
                "break_item__before_period",
            )
        )


class SupervisionSubstitutionManager(CurrentSiteManager):
    """Manager adding specific methods to supervision substitutions."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return (
            super()
            .get_queryset()
            .select_related(
                "teacher",
                "supervision",
                "supervision__teacher",
                "supervision__area",
                "supervision__break_item",
                "supervision__break_item__after_period",
                "supervision__break_item__before_period",
            )
        )


class EventManager(CurrentSiteManager):
    """Manager adding specific methods to events."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return (
            super()
            .get_queryset()
            .select_related("period_from", "period_to")
            .prefetch_related(
                "groups",
                "groups__school_term",
                "groups__parent_groups",
                "teachers",
                "rooms",
            )
        )


class ExtraLessonManager(CurrentSiteManager):
    """Manager adding specific methods to extra lessons."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return (
            super()
            .get_queryset()
            .select_related("room", "period", "subject")
            .prefetch_related(
                "groups",
                "groups__school_term",
                "groups__parent_groups",
                "teachers",
            )
        )


class BreakManager(CurrentSiteManager):
    """Manager adding specific methods to breaks."""

    def get_queryset(self):
        """Ensure all related data is loaded as well."""
        return super().get_queryset().select_related("before_period", "after_period")


class WeekQuerySetMixin:
    def annotate_week(self, week: Union[CalendarWeek]):
        """Annotate all lessons in the QuerySet with the number of the provided calendar week."""
        return self.annotate(
            _week=models.Value(week.week, models.IntegerField()),
            _year=models.Value(week.year, models.IntegerField()),
        )

    def alias_week(self, week: Union[CalendarWeek]):
        """Add an alias to all lessons in the QuerySet with the number of the calendar week."""
        return self.alias(
            _week=models.Value(week.week, models.IntegerField()),
            _year=models.Value(week.year, models.IntegerField()),
        )


class GroupByPeriodsMixin:
    def group_by_periods(self, is_week: bool = False) -> dict:
        """Group a QuerySet of objects with attribute period by period numbers and weekdays."""
        per_period = {}
        for obj in self:
            period = obj.period.period
            weekday = obj.period.weekday

            if period not in per_period:
                per_period[period] = [] if not is_week else {}

            if is_week and weekday not in per_period[period]:
                per_period[period][weekday] = []

            if not is_week:
                per_period[period].append(obj)
            else:
                per_period[period][weekday].append(obj)

        return per_period


class LessonDataQuerySet(models.QuerySet, WeekQuerySetMixin):
    """Overrides default QuerySet to add specific methods for lesson data."""

    # Overridden in the subclasses. Swaps the paths to the base lesson period
    # and to any substitutions depending on whether the query is run on a
    # lesson period or a substitution
    _period_path = None
    _subst_path = None

    def within_dates(self, start: date, end: date):
        """Filter for all lessons within a date range."""
        return self.filter(
            **{
                self._period_path + "lesson__validity__date_start__lte": start,
                self._period_path + "lesson__validity__date_end__gte": end,
            }
        )

    def in_week(self, wanted_week: CalendarWeek):
        """Filter for all lessons within a calendar week."""
        return self.within_dates(
            wanted_week[0] + timedelta(days=1) * (F(self._period_path + "period__weekday")),
            wanted_week[0] + timedelta(days=1) * (F(self._period_path + "period__weekday")),
        ).annotate_week(wanted_week)

    def on_day(self, day: date):
        """Filter for all lessons on a certain day."""
        week, weekday = week_weekday_from_date(day)

        return (
            self.within_dates(day, day)
            .filter(**{self._period_path + "period__weekday": weekday})
            .annotate_week(week)
        )

    def at_time(self, when: Optional[datetime] = None):
        """Filter for the lessons taking place at a certain point in time."""
        now = when or datetime.now()
        week, weekday = week_weekday_from_date(now.date())

        return self.filter(
            **{
                self._period_path + "lesson__validity__date_start__lte": now.date(),
                self._period_path + "lesson__validity__date_end__gte": now.date(),
                self._period_path + "period__weekday": now.weekday(),
                self._period_path + "period__time_start__lte": now.time(),
                self._period_path + "period__time_end__gte": now.time(),
            }
        ).annotate_week(week)

    def filter_participant(self, person: Union[Person, int]):
        """Filter for all lessons a participant (student) attends."""
        return self.filter(Q(**{self._period_path + "lesson__groups__members": person}))

    def filter_group(self, group: Union[Group, int]):
        """Filter for all lessons a group (class) regularly attends."""
        if isinstance(group, int):
            group = Group.objects.get(pk=group)

        if group.parent_groups.all():
            # Prevent to show lessons multiple times
            return self.filter(Q(**{self._period_path + "lesson__groups": group}))
        else:
            return self.filter(
                Q(**{self._period_path + "lesson__groups": group})
                | Q(**{self._period_path + "lesson__groups__parent_groups": group})
            )

    def filter_groups(self, groups: Iterable[Group]) -> QuerySet:
        """Filter for all lessons one of the groups regularly attends."""
        return self.filter(
            Q(**{self._period_path + "lesson__groups__in": groups})
            | Q(**{self._period_path + "lesson__groups__parent_groups__in": groups})
        )

    def filter_teacher(self, teacher: Union[Person, int], is_smart: bool = True):
        """Filter for all lessons given by a certain teacher."""
        qs1 = self.filter(**{self._period_path + "lesson__teachers": teacher})
        qs2 = self.filter(
            **{
                self._subst_path + "teachers": teacher,
                self._subst_path + "week": F("_week"),
                self._subst_path + "year": F("_year"),
            }
        )

        if is_smart:
            return qs1.union(qs2)
        else:
            return qs1

    def filter_room(self, room: Union["Room", int], is_smart: bool = True):
        """Filter for all lessons taking part in a certain room."""
        qs1 = self.filter(**{self._period_path + "room": room})
        qs2 = self.filter(
            **{
                self._subst_path + "room": room,
                self._subst_path + "week": F("_week"),
                self._subst_path + "year": F("_year"),
            }
        )

        if is_smart:
            return qs1.union(qs2)
        else:
            return qs1

    def filter_from_type(
        self, type_: TimetableType, obj: Union[Person, Group, "Room", int], is_smart: bool = True
    ) -> Optional[models.QuerySet]:
        """Filter lesson data for a group, teacher or room by provided type."""
        if type_ == TimetableType.GROUP:
            return self.filter_group(obj)
        elif type_ == TimetableType.TEACHER:
            return self.filter_teacher(obj, is_smart=is_smart)
        elif type_ == TimetableType.ROOM:
            return self.filter_room(obj, is_smart=is_smart)
        else:
            return None

    def filter_from_person(self, person: Person) -> Optional[models.QuerySet]:
        """Filter lesson data for a person."""
        type_ = person.timetable_type

        if type_ == TimetableType.TEACHER:
            # Teacher

            return self.filter_teacher(person)

        elif type_ == TimetableType.GROUP:
            # Student

            return self.filter_participant(person)

        else:
            # If no student or teacher
            return None

    def daily_lessons_for_person(
        self, person: Person, wanted_day: date
    ) -> Optional[models.QuerySet]:
        """Filter lesson data on a day by a person."""
        if person.timetable_type is None:
            return None

        lesson_periods = self.on_day(wanted_day).filter_from_person(person)

        return lesson_periods

    def group_by_validity(self) -> Dict["ValidityRange", List["LessonPeriod"]]:
        """Group lesson periods by validity range as dictionary."""
        lesson_periods_by_validity = {}
        for lesson_period in self:
            lesson_periods_by_validity.setdefault(lesson_period.lesson.validity, [])
            lesson_periods_by_validity[lesson_period.lesson.validity].append(lesson_period)
        return lesson_periods_by_validity

    def next_lesson(
        self, reference: "LessonPeriod", offset: Optional[int] = 1
    ) -> Optional["LessonPeriod"]:
        """Get another lesson in an ordered set of lessons.

        By default, it returns the next lesson in the set. By passing the offset argument,
        the n-th next lesson can be selected. By passing a negative number, the n-th
        previous lesson can be selected.

        This function will handle week, year and validity range changes automatically
        if the queryset contains enough lesson data.
        """
        # Group lesson periods by validity to handle validity range changes correctly
        lesson_periods_by_validity = self.group_by_validity()
        validity_ranges = list(lesson_periods_by_validity.keys())

        # List with lesson periods in the validity range of the reference lesson period
        current_lesson_periods = lesson_periods_by_validity[reference.lesson.validity]
        pks = [lesson_period.pk for lesson_period in current_lesson_periods]

        # Position of the reference lesson period
        index = pks.index(reference.id)

        next_index = index + offset
        if next_index > len(pks) - 1:
            next_index %= len(pks)
            week = reference._week + 1
        elif next_index < 0:
            next_index = len(pks) + next_index
            week = reference._week - 1
        else:
            week = reference._week

        # Check if selected week makes a year change necessary
        year = reference._year
        if week < 1:
            year -= 1
            week = CalendarWeek.get_last_week_of_year(year).week
        elif week > CalendarWeek.get_last_week_of_year(year).week:
            year += 1
            week = 1

        # Get the next lesson period in this validity range and it's date
        # to check whether the validity range has to be changed
        week = CalendarWeek(week=week, year=year)
        next_lesson_period = current_lesson_periods[next_index]
        next_lesson_period_date = week_weekday_to_date(week, next_lesson_period.period.weekday)

        validity_index = validity_ranges.index(next_lesson_period.lesson.validity)

        # If date of next lesson period is out of validity range (smaller) ...
        if next_lesson_period_date < next_lesson_period.lesson.validity.date_start:
            # ... we have to get the lesson period from the previous validity range
            if validity_index == 0:
                # There are no validity ranges (and thus no lessons)
                # in the school term before this lesson period
                return None

            # Get new validity range and last lesson period of this validity range
            new_validity = validity_ranges[validity_index - 1]
            next_lesson_period = lesson_periods_by_validity[new_validity][-1]

            # Build new week with the date from the new validity range/lesson period
            week = CalendarWeek(
                week=new_validity.date_end.isocalendar()[1], year=new_validity.date_end.year
            )

        # If date of next lesson period is out of validity range (larger) ...
        elif next_lesson_period_date > next_lesson_period.lesson.validity.date_end:
            # ... we have to get the lesson period from the next validity range
            if validity_index >= len(validity_ranges) - 1:
                # There are no validity ranges (and thus no lessons)
                # in the school term after this lesson period
                return None

            # Get new validity range and first lesson period of this validity range
            new_validity = validity_ranges[validity_index + 1]
            next_lesson_period = lesson_periods_by_validity[new_validity][0]

            # Build new week with the date from the new validity range/lesson period
            week = CalendarWeek(
                week=new_validity.date_start.isocalendar()[1], year=new_validity.date_start.year
            )

        # Do a new query here to be able to annotate the new week
        return self.annotate_week(week).get(pk=next_lesson_period.pk)


class LessonPeriodQuerySet(LessonDataQuerySet, GroupByPeriodsMixin):
    """QuerySet with custom query methods for lesson periods."""

    _period_path = ""
    _subst_path = "substitutions__"


class LessonSubstitutionQuerySet(LessonDataQuerySet):
    """QuerySet with custom query methods for substitutions."""

    _period_path = "lesson_period__"
    _subst_path = ""

    def within_dates(self, start: date, end: date):
        """Filter for all substitutions within a date range."""
        start_week = CalendarWeek.from_date(start)
        end_week = CalendarWeek.from_date(end)
        return self.filter(
            week__gte=start_week.week,
            week__lte=end_week.week,
            year__gte=start_week.year,
            year__lte=end_week.year,
        ).filter(
            Q(
                week=start_week.week,
                year=start_week.year,
                lesson_period__period__weekday__gte=start.weekday(),
            )
            | Q(
                week=end_week.week,
                year=end_week.year,
                lesson_period__period__weekday__lte=end.weekday(),
            )
            | (
                ~Q(week=start_week.week, year=start_week.year)
                & ~Q(week=end_week.week, year=end_week.year)
            )
        )

    def in_week(self, wanted_week: CalendarWeek):
        """Filter for all lessons within a calendar week."""
        return self.filter(week=wanted_week.week, year=wanted_week.year).annotate_week(wanted_week)

    def on_day(self, day: date):
        """Filter for all lessons on a certain day."""
        week, weekday = week_weekday_from_date(day)

        return self.in_week(week).filter(lesson_period__period__weekday=weekday)

    def at_time(self, when: Optional[datetime] = None):
        """Filter for the lessons taking place at a certain point in time."""
        now = when or datetime.now()

        return self.on_day(now.date()).filter(
            lesson_period__period__time_start__lte=now.time(),
            lesson_period__period__time_end__gte=now.time(),
        )

    def affected_lessons(self):
        """Return all lessons which are affected by selected substitutions."""
        from .models import Lesson  # noaq

        return Lesson.objects.filter(lesson_periods__substitutions__in=self).distinct()

    def affected_teachers(self):
        """Get affected teachers.

        Return all teachers which are affected by
        selected substitutions (as substituted or substituting).
        """
        return (
            Person.objects.filter(
                Q(lessons_as_teacher__in=self.affected_lessons()) | Q(lesson_substitutions__in=self)
            )
            .distinct()
            .order_by("short_name")
        )

    def affected_groups(self):
        """Return all groups which are affected by selected substitutions."""
        return (
            Group.objects.filter(lessons__in=self.affected_lessons())
            .distinct()
            .order_by("short_name")
        )


class DateRangeQuerySetMixin:
    """QuerySet with custom query methods for models with date and period ranges.

    Filterable fields: date_start, date_end, period_from, period_to
    """

    def within_dates(self, start: date, end: date):
        """Filter for all events within a date range."""
        return self.filter(date_start__lte=end, date_end__gte=start)

    def in_week(self, wanted_week: CalendarWeek):
        """Filter for all events within a calendar week."""
        return self.within_dates(wanted_week[0], wanted_week[6])

    def on_day(self, day: date):
        """Filter for all events on a certain day."""
        return self.within_dates(day, day)

    def at_time(self, when: Optional[datetime] = None):
        """Filter for the events taking place at a certain point in time."""
        now = when or datetime.now()

        return self.on_day(now.date()).filter(
            period_from__time_start__lte=now.time(), period_to__time_end__gte=now.time()
        )

    def exclude_holidays(self, holidays: Iterable["Holiday"]) -> QuerySet:
        """Exclude all objects which are in the provided holidays."""
        q = Q()
        for holiday in holidays:
            q = q | Q(date_start__lte=holiday.date_end, date_end__gte=holiday.date_start)
        return self.exclude(q)


class AbsenceQuerySet(DateRangeQuerySetMixin, SchoolTermRelatedQuerySet):
    """QuerySet with custom query methods for absences."""

    def absent_teachers(self):
        return Person.objects.filter(absences__in=self).distinct().order_by("short_name")

    def absent_groups(self):
        return Group.objects.filter(absences__in=self).distinct().order_by("short_name")

    def absent_rooms(self):
        return Person.objects.filter(absences__in=self).distinct().order_by("short_name")


class HolidayQuerySet(QuerySet, DateRangeQuerySetMixin):
    """QuerySet with custom query methods for holidays."""

    def get_all_days(self) -> List[date]:
        """Get all days included in the selected holidays."""
        holiday_days = []
        for holiday in self:
            holiday_days += list(holiday.get_days())
        return holiday_days


class SupervisionQuerySet(ValidityRangeRelatedQuerySet, WeekQuerySetMixin):
    """QuerySet with custom query methods for supervisions."""

    def filter_by_weekday(self, weekday: int):
        """Filter supervisions by weekday."""
        return self.filter(
            Q(break_item__before_period__weekday=weekday)
            | Q(break_item__after_period__weekday=weekday)
        )

    def filter_by_teacher(self, teacher: Union[Person, int]):
        """Filter for all supervisions given by a certain teacher."""
        if self.count() > 0:
            if hasattr(self[0], "_week"):
                week = CalendarWeek(week=self[0]._week, year=self[0]._year)
            else:
                week = CalendarWeek.current_week()

            dates = [week[w] for w in range(0, 7)]

            return self.filter(
                Q(substitutions__teacher=teacher, substitutions__date__in=dates)
                | Q(teacher=teacher)
            )

        return self


class TimetableQuerySet(models.QuerySet):
    """Common query set methods for objects in timetables.

    Models need following fields:
    - groups
    - teachers
    - rooms (_multiple_rooms=True)/room (_multiple_rooms=False)
    """

    _multiple_rooms = True

    def filter_participant(self, person: Union[Person, int]):
        """Filter for all objects a participant (student) attends."""
        return self.filter(Q(groups__members=person))

    def filter_group(self, group: Union[Group, int]):
        """Filter for all objects a group (class) attends."""
        if isinstance(group, int):
            group = Group.objects.get(pk=group)

        if group.parent_groups.all():
            # Prevent to show lessons multiple times
            return self.filter(groups=group)
        else:
            return self.filter(Q(groups=group) | Q(groups__parent_groups=group))

    def filter_groups(self, groups: Iterable[Group]) -> QuerySet:
        """Filter for all objects one of the groups attends."""
        return self.filter(Q(groups__in=groups) | Q(groups__parent_groups__in=groups)).distinct()

    def filter_teacher(self, teacher: Union[Person, int]):
        """Filter for all lessons given by a certain teacher."""
        return self.filter(teachers=teacher)

    def filter_room(self, room: Union["Room", int]):
        """Filter for all objects taking part in a certain room."""
        if self._multiple_rooms:
            return self.filter(rooms=room)
        else:
            return self.filter(room=room)

    def filter_from_type(
        self, type_: TimetableType, obj: Union[Group, Person, "Room", int]
    ) -> Optional[models.QuerySet]:
        """Filter data for a group, teacher or room by provided type."""
        if type_ == TimetableType.GROUP:
            return self.filter_group(obj)
        elif type_ == TimetableType.TEACHER:
            return self.filter_teacher(obj)
        elif type_ == TimetableType.ROOM:
            return self.filter_room(obj)
        else:
            return None

    def filter_from_person(self, person: Person) -> Optional[models.QuerySet]:
        """Filter data by person."""
        type_ = person.timetable_type

        if type_ == TimetableType.TEACHER:
            # Teacher

            return self.filter_teacher(person)

        elif type_ == TimetableType.GROUP:
            # Student

            return self.filter_participant(person)

        else:
            # If no student or teacher
            return None


class EventQuerySet(DateRangeQuerySetMixin, SchoolTermRelatedQuerySet, TimetableQuerySet):
    """QuerySet with custom query methods for events."""

    def annotate_day(self, day: date):
        """Annotate all events in the QuerySet with the provided date."""
        return self.annotate(_date=models.Value(day, models.DateField()))

    def alias_day(self, day: date):
        """Add an alias to all events in the QuerySet with the provided date."""
        return self.alias(_date=models.Value(day, models.DateField()))


class ExtraLessonQuerySet(TimetableQuerySet, SchoolTermRelatedQuerySet, GroupByPeriodsMixin):
    """QuerySet with custom query methods for extra lessons."""

    _multiple_rooms = False

    def within_dates(self, start: date, end: date):
        """Filter all extra lessons within a specific time range."""
        return self.alias_day().filter(day__gte=start, day__lte=end)

    def on_day(self, day: date):
        """Filter all extra lessons on a day."""
        return self.within_dates(day, day)

    def _get_weekday_to_date(self):
        """Get DB function to convert a weekday to a date."""
        return ExpressionWrapper(
            Func(
                Concat(F("year"), F("week")),
                Value("IYYYIW"),
                output_field=DateField(),
                function="TO_DATE",
            )
            + F("period__weekday"),
            output_field=DateField(),
        )

    def annotate_day(self):
        return self.annotate(day=self._get_weekday_to_date())

    def alias_day(self):
        return self.alias(day=self._get_weekday_to_date())

    def exclude_holidays(self, holidays: Iterable["Holiday"]) -> QuerySet:
        """Exclude all extra lessons which are in the provided holidays."""
        q = Q()
        for holiday in holidays:
            q = q | Q(day__lte=holiday.date_end, day__gte=holiday.date_start)
        return self.alias_day().exclude(q)


class GroupPropertiesMixin:
    """Mixin for common group properties.

    Needed field: `groups`
    """

    @property
    def group_names(self, sep: Optional[str] = ", ") -> str:
        return sep.join([group.short_name for group in self.groups.all()])

    @property
    def groups_to_show(self) -> models.QuerySet:
        groups = self.groups.all()
        if (
            groups.count() == 1
            and groups[0].parent_groups.all()
            and get_site_preferences()["chronos__use_parent_groups"]
        ):
            return groups[0].parent_groups.all()
        else:
            return groups

    @property
    def groups_to_show_names(self, sep: Optional[str] = ", ") -> str:
        return sep.join([group.short_name for group in self.groups_to_show])


class TeacherPropertiesMixin:
    """Mixin for common teacher properties.

    Needed field: `teacher`
    """

    @property
    def teacher_names(self, sep: Optional[str] = ", ") -> str:
        return sep.join([teacher.full_name for teacher in self.get_teachers().all()])

    @property
    def teacher_short_names(self, sep: str = ", ") -> str:
        return sep.join([teacher.short_name for teacher in self.get_teachers().all()])



# Generated by Django 3.2.9 on 2021-12-20 16:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('chronos', '0009_automaticplan'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='subject',
            name='unique_name_per_site',
        ),
    ]



# Generated by Django 3.0.7 on 2020-06-27 14:00

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.db import migrations, models


def migrate_lesson(apps, schema_editor):
    Lesson = apps.get_model("chronos", "Lesson")
    ValidityRange = apps.get_model("chronos", "ValidityRange")
    SchoolTerm = apps.get_model("core", "SchoolTerm")

    db_alias = schema_editor.connection.alias

    for lesson in Lesson.objects.using(db_alias).all():
        date_start = lesson.date_start
        date_end = lesson.date_end

        try:
            school_term = SchoolTerm.objects.using(db_alias).get(
                date_start__lte=date_end, date_end__gte=date_start
            )
        except SchoolTerm.DoesNotExist:
            school_term = SchoolTerm.objects.using(db_alias).create(
                date_start=date_start, date_end=date_end
            )

        try:
            validity_range = ValidityRange.objects.using(db_alias).get(
                school_term=school_term,
                date_start__lte=date_end,
                date_end__gte=date_start,
            )
        except ValidityRange.DoesNotExist:
            validity_range = ValidityRange.objects.using(db_alias).create(
                date_start=date_start, date_end=date_end, school_term=school_term
            )

        lesson.validity = validity_range
        lesson.save()


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0002_school_term"),
        ("sites", "0002_alter_domain_unique"),
        ("chronos", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ValidityRange",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "extended_data",
                    models.JSONField(
                        default=dict, editable=False
                    ),
                ),
                (
                    "name",
                    models.CharField(blank=True, max_length=255, verbose_name="Name"),
                ),
                ("date_start", models.DateField(verbose_name="Start date")),
                ("date_end", models.DateField(verbose_name="End date")),
            ],
            options={
                "verbose_name": "Validity range",
                "verbose_name_plural": "Validity ranges",
            },
        ),
        migrations.AlterModelOptions(
            name="lesson",
            options={
                "ordering": ["validity__date_start", "subject"],
                "verbose_name": "Lesson",
                "verbose_name_plural": "Lessons",
            },
        ),
        migrations.AlterModelOptions(
            name="lessonperiod",
            options={
                "ordering": [
                    "lesson__validity__date_start",
                    "period__weekday",
                    "period__period",
                    "lesson__subject",
                ],
                "verbose_name": "Lesson period",
                "verbose_name_plural": "Lesson periods",
            },
        ),
        migrations.AlterModelOptions(
            name="lessonsubstitution",
            options={
                "ordering": [
                    "lesson_period__lesson__validity__date_start",
                    "week",
                    "lesson_period__period__weekday",
                    "lesson_period__period__period",
                ],
                "verbose_name": "Lesson substitution",
                "verbose_name_plural": "Lesson substitutions",
            },
        ),
        migrations.AlterModelManagers(name="break", managers=[],),
        migrations.AlterModelManagers(name="exam", managers=[],),
        migrations.AlterModelManagers(name="lesson", managers=[],),
        migrations.AlterModelManagers(name="timeperiod", managers=[],),
        migrations.AddField(
            model_name="absence",
            name="school_term",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.SchoolTerm",
                verbose_name="Linked school term",
            ),
        ),
        migrations.AddField(
            model_name="break",
            name="school_term",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.SchoolTerm",
                verbose_name="Linked school term",
            ),
        ),
        migrations.AddField(
            model_name="event",
            name="school_term",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.SchoolTerm",
                verbose_name="Linked school term",
            ),
        ),
        migrations.AddField(
            model_name="exam",
            name="school_term",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.SchoolTerm",
                verbose_name="Linked school term",
            ),
        ),
        migrations.AddField(
            model_name="extralesson",
            name="school_term",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.SchoolTerm",
                verbose_name="Linked school term",
            ),
        ),
        migrations.AddField(
            model_name="validityrange",
            name="school_term",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="validity_ranges",
                to="core.SchoolTerm",
                verbose_name="School term",
            ),
        ),
        migrations.AddField(
            model_name="validityrange",
            name="site",
            field=models.ForeignKey(
                default=1,
                editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="lesson",
            name="validity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="chronos.ValidityRange",
                verbose_name="Linked validity range",
                null=True,
                blank=True,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="supervision",
            name="validity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="chronos.ValidityRange",
                verbose_name="Linked validity range",
                null=True,
                blank=True,
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="timeperiod",
            name="validity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="chronos.ValidityRange",
                verbose_name="Linked validity range",
                null=True,
                blank=True,
            ),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name="validityrange", unique_together={("date_start", "date_end")},
        ),
        migrations.RunPython(migrate_lesson),
        migrations.RemoveIndex(
            model_name="lesson", name="chronos_les_date_st_5ecc62_idx",
        ),
        migrations.RemoveField(model_name="lesson", name="date_end",),
        migrations.RemoveField(model_name="lesson", name="date_start",),
    ]



from django.contrib.auth.models import User
from django.db.models import Model

from rules import predicate

from aleksis.core.models import Group, Person
from aleksis.core.util.predicates import has_global_perm, has_object_perm

from ..models import Room
from .chronos_helpers import get_classes, get_rooms, get_teachers


@predicate
def has_timetable_perm(user: User, obj: Model) -> bool:
    """
    Check if can access timetable.

    Predicate which checks whether the user is allowed
    to access the requested timetable.
    """
    if isinstance(obj, Group):
        return has_group_timetable_perm(user, obj)
    elif isinstance(obj, Person):
        return has_person_timetable_perm(user, obj)
    elif isinstance(obj, Room):
        return has_room_timetable_perm(user, obj)
    else:
        return False


@predicate
def has_group_timetable_perm(user: User, obj: Group) -> bool:
    """
    Check if can access group timetable.

    Predicate which checks whether the user is allowed
    to access the requested group timetable.
    """
    return (
        obj in user.person.member_of.all()
        or user.person.primary_group == obj
        or obj in user.person.owner_of.all()
        or has_global_perm("chronos.view_all_group_timetables")(user)
        or has_object_perm("core.view_group_timetable")(user, obj)
    )


@predicate
def has_person_timetable_perm(user: User, obj: Person) -> bool:
    """
    Check if can access person timetable.

    Predicate which checks whether the user is allowed
    to access the requested person timetable.
    """
    return (
        user.person == obj
        or has_global_perm("chronos.view_all_person_timetables")(user)
        or has_object_perm("core.view_person_timetable")(user, obj)
    )


@predicate
def has_room_timetable_perm(user: User, obj: Room) -> bool:
    """
    Check if can access room timetable.

    Predicate which checks whether the user is allowed
    to access the requested room timetable.
    """
    return has_global_perm("chronos.view_all_room_timetables")(user) or has_object_perm(
        "chronos.view_room_timetable"
    )(user, obj)


@predicate
def has_any_timetable_object(user: User) -> bool:
    """Predicate which checks whether there are any timetables the user is allowed to access."""
    return get_classes(user).exists() or get_rooms(user).exists() or get_teachers(user).exists()



# Generated by Django 3.0.7 on 2020-06-30 10:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("chronos", "0002_school_term_validity"),
    ]

    operations = [
        migrations.RemoveField(model_name="break", name="school_term",),
        migrations.AddField(
            model_name="break",
            name="validity",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="chronos.ValidityRange",
                verbose_name="Linked validity range",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="timeperiod", unique_together={("weekday", "period", "validity")},
        ),
    ]



from datetime import date
from typing import Union

from django.db import models
from django.utils.translation import gettext as _

from calendarweek import CalendarWeek

from aleksis.apps.chronos.util.date import week_weekday_to_date
from aleksis.core.managers import CurrentSiteManagerWithoutMigrations
from aleksis.core.mixins import ExtensibleModel

from .managers import ValidityRangeRelatedQuerySet


class ValidityRangeRelatedExtensibleModel(ExtensibleModel):
    """Add relation to validity range."""

    objects = CurrentSiteManagerWithoutMigrations.from_queryset(ValidityRangeRelatedQuerySet)()

    validity = models.ForeignKey(
        "chronos.ValidityRange",
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("Linked validity range"),
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True


class WeekRelatedMixin:
    @property
    def date(self) -> date:
        period = self.lesson_period.period if hasattr(self, "lesson_period") else self.period
        return week_weekday_to_date(self.calendar_week, period.weekday)

    @property
    def calendar_week(self) -> CalendarWeek:
        return CalendarWeek(week=self.week, year=self.year)


class WeekAnnotationMixin:
    def annotate_week(self, week: CalendarWeek):
        """Annotate this lesson with the number of the provided calendar week."""
        self._week = week.week
        self._year = week.year

    @property
    def week(self) -> Union[CalendarWeek, None]:
        """Get annotated week as `CalendarWeek`.

        Defaults to `None` if no week is annotated.
        """
        if hasattr(self, "_week"):
            return CalendarWeek(week=self._week, year=self._year)
        else:
            return None



from datetime import time, timedelta

from django.contrib.auth import get_user_model
from django.utils import timezone

import pytest

from aleksis.apps.chronos.util.chronos_helpers import get_rooms, get_teachers
from aleksis.core.models import Group, Person, SchoolTerm

pytestmark = pytest.mark.django_db


from aleksis.apps.chronos.models import (
    Lesson,
    LessonPeriod,
    Room,
    Subject,
    TimePeriod,
    ValidityRange,
)


def test_rooms_teachers_only_from_current_school_term():
    User = get_user_model()

    user = User.objects.create(username="test", is_staff=True, is_superuser=True)
    person_user = Person.objects.create(user=user, first_name="Test", last_name="User")

    correct_school_term = SchoolTerm.objects.create(
        date_start=timezone.now() - timedelta(days=1),
        date_end=timezone.now() + timedelta(days=1),
        name="Correct school term",
    )
    wrong_school_term = SchoolTerm.objects.create(
        date_start=timezone.now() - timedelta(days=3),
        date_end=timezone.now() - timedelta(days=2),
        name="Wrong school term",
    )

    correct_validity = ValidityRange.objects.create(
        school_term=correct_school_term,
        date_start=correct_school_term.date_start,
        date_end=correct_school_term.date_end,
        name="Correct validity",
    )
    wrong_validity = ValidityRange.objects.create(
        school_term=wrong_school_term,
        date_start=wrong_school_term.date_start,
        date_end=wrong_school_term.date_end,
        name="Wrong validity",
    )

    subject = Subject.objects.create(name="Test subject", short_name="TS")
    time_period = TimePeriod.objects.create(
        weekday=0, period=1, time_start=time(8, 0), time_end=time(9, 0)
    )

    correct_person = Person.objects.create(first_name="Correct", last_name="Person")
    wrong_person = Person.objects.create(first_name="Wrong", last_name="Person")

    correct_lesson = Lesson.objects.create(validity=correct_validity, subject=subject)
    correct_lesson.teachers.add(correct_person)
    wrong_lesson = Lesson.objects.create(validity=wrong_validity, subject=subject)
    wrong_lesson.teachers.add(wrong_person)

    correct_room = Room.objects.create(name="Correct room", short_name="cr")
    wrong_room = Room.objects.create(name="Wrong room", short_name="wr")

    correct_lesson_period = LessonPeriod.objects.create(
        lesson=correct_lesson, period=time_period, room=correct_room
    )
    wrong_lesson_period = LessonPeriod.objects.create(
        lesson=wrong_lesson, period=time_period, room=wrong_room
    )

    rooms = get_rooms(user)
    assert correct_room in rooms
    assert wrong_room not in rooms

    teachers = get_teachers(user)
    assert correct_person in teachers
    assert wrong_person not in teachers



# Generated by Django 3.2.13 on 2022-11-09 23:07

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('chronos', '0011_exam'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='chronosglobalpermissions',
            options={'managed': False, 'permissions': (('view_all_room_timetables', 'Can view all room timetables'), ('view_all_group_timetables', 'Can view all group timetables'), ('view_all_person_timetables', 'Can view all person timetables'), ('view_timetable_overview', 'Can view timetable overview'), ('view_lessons_day', 'Can view all lessons per day'), ('view_supervisions_day', 'Can view all supervisions per day'))},
        ),
    ]



from datetime import date, datetime
from typing import Optional, Union

from django import template
from django.db.models.query import QuerySet

from aleksis.apps.chronos.util.date import CalendarWeek, week_period_to_date, week_weekday_to_date

register = template.Library()


@register.filter
def week_start(week: CalendarWeek) -> date:
    return week[0]


@register.filter
def week_end(week: CalendarWeek) -> date:
    return week[-1]


@register.filter
def only_week(qs: QuerySet, week: Optional[CalendarWeek]) -> QuerySet:
    wanted_week = week or CalendarWeek()
    return qs.filter(week=wanted_week.week, year=wanted_week.year)


@register.simple_tag
def weekday_to_date(week: CalendarWeek, weekday: int) -> date:
    return week_weekday_to_date(week, weekday)


@register.simple_tag
def period_to_date(week: CalendarWeek, period) -> date:
    return week_period_to_date(week, period)


@register.simple_tag
def period_to_time_start(date_ref: Union[CalendarWeek, int, date], period) -> date:
    return period.get_datetime_start(date_ref)


@register.simple_tag
def period_to_time_end(date_ref: Union[CalendarWeek, int, date], period) -> date:
    return period.get_datetime_end(date_ref)


@register.simple_tag
def today() -> date:
    return date.today()


@register.simple_tag
def now_datetime() -> datetime:
    return datetime.now()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aleksis',
 'aleksis.apps.chronos',
 'aleksis.apps.chronos.migrations',
 'aleksis.apps.chronos.templatetags',
 'aleksis.apps.chronos.tests',
 'aleksis.apps.chronos.tests.regression',
 'aleksis.apps.chronos.util']

package_data = \
{'': ['*'],
 'aleksis.apps.chronos': ['locale/ar/LC_MESSAGES/*',
                          'locale/de_DE/LC_MESSAGES/*',
                          'locale/fr/LC_MESSAGES/*',
                          'locale/la/LC_MESSAGES/*',
                          'locale/nb_NO/LC_MESSAGES/*',
                          'locale/ru/LC_MESSAGES/*',
                          'locale/tr_TR/LC_MESSAGES/*',
                          'locale/uk/LC_MESSAGES/*',
                          'static/css/chronos/*',
                          'static/js/chronos/*',
                          'templates/chronos/*',
                          'templates/chronos/partials/*',
                          'templates/chronos/partials/subs/*',
                          'templates/search/indexes/chronos/*']}

install_requires = \
['aleksis-app-resint>=2.0,<3.0',
 'aleksis-core>=2.8,<3.0',
 'calendarweek>=0.5.0,<0.6.0']

entry_points = \
{'aleksis.app': ['chronos = aleksis.apps.chronos.apps:ChronosConfig']}

setup_kwargs = {
    'name': 'aleksis-app-chronos',
    'version': '2.5',
    'description': 'AlekSIS (School Information System)\u200a—\u200aApp Χρόνος (digital timetables)',
    'long_description': 'AlekSIS (School Information System)\u200a—\u200aApp Χρόνος (digital timetables)\n=====================================================================\n\nAlekSIS\n-------\n\nThis is an application for use with the `AlekSIS®`_ platform.\n\nFeatures\n--------\n\n* Consider object-level rules and permissions\n* Export timetables as PDF\n* Search in rooms\n* Show absent groups in timetable\n* Show absent teachers in timetable\n* Show affected groups in timetable\n* Show affected teachers in timetable\n* Show supervisions in timetable\n* Timetables per day\n* Timetables per group\n* Timetables per person\n* Timetables per room\n* Timetables per week\n* Smart timetable\n\nLicence\n-------\n\n::\n\n  Copyright © 2018, 2019, 2020, 2021, 2022 Jonathan Weth <dev@jonathanweth.de>\n  Copyright © 2018, 2019 Frank Poetzsch-Heffter <p-h@katharineum.de>\n  Copyright © 2019, 2020, 2022 Dominik George <dominik.george@teckids.org>\n  Copyright © 2019, 2021 Hangzhi Yu <yuha@katharineum.de>\n  Copyright © 2019 Julian Leucker <leuckeju@katharineum.de>\n  Copyright © 2019 Tom Teichler <tom.teichler@teckids.org>\n  Copyright © 2021 Lloyd Meins <meinsll@katharineum.de>\n\n  Licenced under the EUPL, version 1.2 or later, by Teckids e.V. (Bonn, Germany).\n\nPlease see the LICENCE.rst file accompanying this distribution for the\nfull licence text or on the `European Union Public Licence`_ website\nhttps://joinup.ec.europa.eu/collection/eupl/guidelines-users-and-developers\n(including all other official language versions).\n\nTrademark\n---------\n\nAlekSIS® is a registered trademark of the AlekSIS open source project, represented\nby Teckids e.V. Please refer to the `trademark policy`_ for hints on using the trademark\nAlekSIS®.\n\n.. _AlekSIS®: https://aleksis.org/\n.. _European Union Public Licence: https://eupl.eu/\n.. _trademark policy: https://aleksis.org/pages/about\n',
    'author': 'Dominik George',
    'author_email': 'dominik.george@teckids.org',
    'maintainer': 'Jonathan Weth',
    'maintainer_email': 'wethjo@katharineum.de',
    'url': 'https://aleksis.org/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Generated by Django 3.1.3 on 2020-12-22 12:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('chronos', '0004_substitution_extra_lesson_year'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='room',
            options={'ordering': ['name', 'short_name'],
                     'permissions': (('view_room_timetable', 'Can view room timetable'),),
                     'verbose_name': 'Room', 'verbose_name_plural': 'Rooms'},
        ),
        migrations.AlterModelOptions(
            name='chronosglobalpermissions',
            options={'managed': False, 'permissions': (('view_all_room_timetables', 'Can view all room timetables'), ('view_all_group_timetables', 'Can view all group timetables'), ('view_all_person_timetables', 'Can view all person timetables'), ('view_timetable_overview', 'Can view timetable overview'), ('view_lessons_day', 'Can view all lessons per day'))},
        ),
    ]



# Generated by Django 3.2.3 on 2021-05-21 18:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('chronos', '0006_indexes'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='chronosglobalpermissions',
            options={'managed': False, 'permissions': (('view_all_room_timetables', 'Can view all room timetables'), ('view_all_group_timetables', 'Can view all group timetables'), ('view_all_person_timetables', 'Can view all person timetables'), ('view_timetable_overview', 'Can view timetable overview'), ('view_lessons_day', 'Can view all lessons per day'))},
        ),
    ]



from __future__ import annotations

from typing import Optional, Union

from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

import django_tables2 as tables
from django_tables2.utils import A, Accessor

from .models import LessonPeriod, Supervision


def _css_class_from_lesson_or_supervision_state(
    record: Optional[Union[LessonPeriod, Supervision]] = None,
    table: Optional[Union[LessonsTable, SupervisionsTable]] = None,
) -> str:
    """Return CSS class depending on lesson or supervision state."""
    if record.get_substitution():
        if hasattr(record.get_substitution(), "cancelled") and record.get_substitution().cancelled:
            return "success"
        else:
            return "warning"
    else:
        return ""


class SubstitutionColumn(tables.Column):
    def render(self, value, record: Optional[Union[LessonPeriod, Supervision]] = None):
        if record.get_substitution():
            return format_html(
                "<s>{}</s> → {}",
                value,
                self.substitution_accessor.resolve(record.get_substitution()),
            )
        return value

    def __init__(self, *args, **kwargs):
        self.substitution_accessor = Accessor(kwargs.pop("substitution_accessor"))
        super().__init__(*args, **kwargs)


class LessonsTable(tables.Table):
    """Table for daily lessons and management of substitutions."""

    class Meta:
        attrs = {"class": "highlight"}
        row_attrs = {"class": _css_class_from_lesson_or_supervision_state}

    period__period = tables.Column(accessor="period__period")
    lesson__groups = tables.Column(accessor="lesson__group_names", verbose_name=_("Groups"))
    lesson__teachers = SubstitutionColumn(
        accessor="lesson__teacher_names",
        substitution_accessor="teacher_names",
        verbose_name=_("Teachers"),
    )
    lesson__subject = SubstitutionColumn(
        accessor="lesson__subject", substitution_accessor="subject"
    )
    room = SubstitutionColumn(accessor="room", substitution_accessor="room")
    edit_substitution = tables.LinkColumn(
        "edit_substitution",
        args=[A("id"), A("_week")],
        text=_("Substitution"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange"}},
        verbose_name=_("Manage substitution"),
    )


class SupervisionsTable(tables.Table):
    """Table for daily supervisions and management of substitutions."""

    class Meta:
        attrs = {"class": "highlight"}
        row_attrs = {"class": _css_class_from_lesson_or_supervision_state}

    break_item = tables.Column(accessor="break_item")
    area = tables.Column(accessor="area")
    teacher = SubstitutionColumn(
        accessor="teacher",
        substitution_accessor="teacher",
        verbose_name=_("Teachers"),
    )
    edit_substitution = tables.LinkColumn(
        "edit_supervision_substitution",
        args=[A("id"), A("_week")],
        text=_("Substitution"),
        attrs={"a": {"class": "btn-flat waves-effect waves-orange"}},
        verbose_name=_("Manage substitution"),
    )
